#pragma once

#include "common/pcl/pcl_types.h"

namespace oh_my_loam {

// 定义点类型的枚举
enum class PointType {
  FLAT_SURF = -2,  // 平面表面
  SURF = -1,       // 曲面
  NORNAL = 0,      // 普通点
  CORNER = 1,      // 角点
  SHARP_CORNER = 2 // 锐角点
};

// 定义点云类型
struct PointXYZT;
using TPoint = PointXYZT;  // 定义一个点类型
using TPointCloud = pcl::PointCloud<TPoint>;  // 点云类型
using TPointCloudPtr = TPointCloud::Ptr;      // 点云指针类型
using TPointCloudConstPtr = TPointCloud::ConstPtr;   // 点云常量指针类型

struct PointXYZTCT;
using TCTPoint = PointXYZTCT;  // 定义带类型和曲率的点类型
using TCTPointCloud = pcl::PointCloud<TCTPoint>;   // 带类型和曲率的点云类型
using TCTPointCloudPtr = TCTPointCloud::Ptr;       // 带类型和曲率的点云指针类型
using TCTPointCloudConstPtr = TCTPointCloud::ConstPtr;  // 带类型和曲率的点云常量指针类型

// 定义PointXYZT结构体，表示具有位置和时间的点
struct EIGEN_ALIGN16 PointXYZT {
  PCL_ADD_POINT4D;   // 添加PCL的4D坐标
  union {
    float time;   // 时间戳
    float intensity;  // 强度值
    float data_c[4];  // 用于存储其他数据
  };

  // 默认构造函数，初始化为零值
  PointXYZT() {
    x = y = z = 0.0f;
    data[3] = 1.0f;
    time = 0.0f;
  }

  // 带参数构造函数
  PointXYZT(float x, float y, float z, float time = 0.0f)
      : x(x), y(y), z(z), time(time) {
    data[3] = 1.0f;
  }

  // 从common::Point构造
  PointXYZT(const common::Point &p) {
    x = p.x;
    y = p.y;
    z = p.z;
    data[3] = 1.0f;
    time = 0.0f;
  }

  // 拷贝构造函数
  PointXYZT(const PointXYZT &p) {
    x = p.x;
    y = p.y;
    z = p.z;
    data[3] = 1.0f;
    time = p.time;
  }
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW  // 确保数据结构的对齐
};

// 定义PointXYZTCT结构体，表示带有类型和曲率信息的点
struct PointXYZTCT {
  PCL_ADD_POINT4D;  // 添加PCL的4D坐标
  union EIGEN_ALIGN16 {
    struct {
      float time;  // 时间戳
      float curvature;  // 曲率
      PointType type;  // 点的类型（如角点、平面点等）
    };
    float data_c[4];  // 用于存储其他数据
  };

  // 默认构造函数，初始化为零值
  PointXYZTCT() {
    x = y = z = 0.0f;
    data[3] = 1.0f;
    time = curvature = 0.0f;
    type = PointType::NORNAL;  // 默认点类型为普通点
  }

  // 带参数构造函数
  PointXYZTCT(float x, float y, float z, float time = 0.0f,
              float curvature = 0.0f, PointType type = PointType::NORNAL)
      : x(x), y(y), z(z), time(time), curvature(curvature), type(type) {
    data[3] = 1.0f;
  }

  // 从common::Point构造
  PointXYZTCT(const common::Point &p) {
    x = p.x;
    y = p.y;
    z = p.z;
    data[3] = 1.0f;
    time = 0.0f;
    curvature = 0.0f;
    type = PointType::NORNAL;
  }

  // 从TPoint构造
  PointXYZTCT(const TPoint &p) {
    x = p.x;
    y = p.y;
    z = p.z;
    data[3] = 1.0f;
    time = p.time;
    curvature = 0.0f;
    type = PointType::NORNAL;
  }

  // 拷贝构造函数
  PointXYZTCT(const PointXYZTCT &p) {
    x = p.x;
    y = p.y;
    z = p.z;
    data[3] = 1.0f;
    time = p.time;
    curvature = p.curvature;
    type = p.type;
  }

  EIGEN_MAKE_ALIGNED_OPERATOR_NEW // 确保数据结构的对齐
} EIGEN_ALIGN16; // 对齐

}  // namespace oh_my_loam

// clang-format off
// 注册PointXYZT类型的点云结构
POINT_CLOUD_REGISTER_POINT_STRUCT(
  oh_my_loam::PointXYZT,
  (float, x, x)
  (float, y, y)
  (float, z, z)
  (float, time, time)
  (float, intensity, intensity)
)

// 注册PointXYZTCT类型的点云结构
POINT_CLOUD_REGISTER_POINT_STRUCT(
  oh_my_loam::PointXYZTCT,
  (float, x, x)
  (float, y, y)
  (float, z, z)
  (float, time, time)
  (float, curvature, curvature)
)
// clang-format on