#pragma once

#include <vector>

#include "common/common.h"
#include "oh_my_loam/base/types.h"

namespace oh_my_loam {

class Row;

class Grid;

// Index 结构体，表示地图中的一个索引位置，包含三个维度（k, j, i）
struct Index {
  int k, j, i;  // 三个维度的索引

  // 默认构造函数
  Index() = default;

  // 构造函数，初始化k, j, i
  Index(int k, int j, int i) : k(k), j(j), i(i) {}

  // 用于排序的比较函数对象，按k、j、i的顺序依次比较
  struct Comp {
    bool operator()(const Index &idx1, const Index &idx2) const {
      return (idx1.k < idx2.k) || (idx1.k == idx2.k && idx1.j < idx2.j) ||
             (idx1.k == idx2.k && idx1.j == idx2.j && idx1.i < idx2.i);
    }
  };
};

// Map 类，表示一个三维网格地图
class Map {
 public:
  // 构造函数，使用地图的形状和步长初始化地图
  Map(const std::vector<int> &shape, const std::vector<double> &step);

  // 构造函数，使用地图的形状和步长初始化地图
  Map(const std::vector<int> &shape, double step);

  // 根据索引返回指定位置的点云
  TPointCloudPtr &at(const Index &index);

  // 根据索引返回指定位置的点云（const版本）
  const TPointCloudPtr &at(const Index &index) const;

  // 清空地图
  void clear();

  // Z轴方向上平移地图n个单位
  void ShiftZ(int n);

  // Y轴方向上平移地图n个单位
  void ShiftY(int n);

  // X轴方向上平移地图n个单位
  void ShiftX(int n);

  // 返回地图的形状
  const std::vector<int> shape() const {
    return std::vector<int>(shape_, shape_ + 3);
  }

  // 根据点的位置计算该点在地图中的索引
  Index GetIndex(const TPoint &point) const;

  // 检查索引是否有效
  bool CheckIndex(const Index &index) const;

  // 获取基于点和子地图形状的点云
  TPointCloudPtr GetSubmapPoints(const TPoint &point,
                                 const std::vector<int> &submap_shapes) const;

  // 获取所有的点云
  TPointCloudPtr GetAllPoints() const;

  // 向地图中添加点云
  void AddPoints(const TPointCloudConstPtr &cloud,
                 std::vector<Index> *const indices = nullptr);

  // 对点云进行下采样
  void Downsample(double voxel_size);

  // 对指定的点云索引进行下采样
  void Downsample(const std::vector<Index> &indices, double voxel_size);

 private:
  // 获取指定的Z轴索引的网格
  Grid &at(int z_idx);

  // 获取指定的Z轴索引的网格（const版本）
  const Grid &at(int z_idx) const;

  // 获取指定Z轴和Y轴索引的行
  Row &at(int z_idx, int y_idx);

  // 获取指定Z轴和Y轴索引的行（const版本）
  const Row &at(int z_idx, int y_idx) const;

  // 获取指定Z轴、Y轴和X轴索引的点云
  TPointCloudPtr &at(int z_idx, int y_idx, int x_idx);

  const TPointCloudPtr &at(int z_idx, int y_idx, int x_idx) const;

  std::vector<Index> GetSubmapIndices(
      const TPoint &point, const std::vector<int> &submap_shapes) const;

  // 获取地图的所有索引
  std::vector<Index> GetAllIndices() const;

  int shape_[3], center_[3];  // 地图的形状和中心位置（z, y, x）

  double step_[3];  // 步长（分辨率）

  std::vector<Grid> map_;  // 存储地图的网格

  DISALLOW_COPY_AND_ASSIGN(Map);  // 禁止拷贝和赋值
};

// Row 类，表示地图中的一行，包含多个点云
class Row {
 public:
  // 构造函数，初始化行的大小
  explicit Row(int n);

  // 根据索引返回指定位置的点云
  TPointCloudPtr &at(int idx);

  // 根据索引返回指定位置的点云（const版本）
  const TPointCloudPtr &at(int idx) const;

  // 清空该行的点云
  void clear();

  // 获取该行的所有点云
  TPointCloudPtr GetAllPoints() const;

 private:
  std::vector<TPointCloudPtr> row_;  // 存储行的点云
};

// Grid 类，表示地图中的一个网格，包含多个行
class Grid {
 public:
  // 构造函数，初始化网格的大小
  Grid(int m, int n);

  // 清空网格
  void clear();

  // 获取指定索引的行
  Row &at(int idx);

  // 获取指定索引的行（const版本）
  const Row &at(int idx) const;

  // 获取该网格的所有点云
  TPointCloudPtr GetAllPoints() const;

 private:
  std::vector<Row> grid_;  // 存储网格中的行
};

}  // namespace oh_my_loam
