#include "oh_my_loam.h"

#include <vector>

#include "common/pcl/pcl_utils.h"
#include "common/registerer/registerer.h"

namespace oh_my_loam {

namespace {
const double kPointMinDist = 0.5; // 点云中点的最小距离阈值
}  // namespace

// 初始化OhMyLoam系统
bool OhMyLoam::Init() {
  // 加载配置文件
  config_ = common::YAMLConfig::Instance()->config();
  is_vis_ = config_["vis"].as<bool>(); // 是否启用可视化
  // 创建提取器实例
  extractor_.reset(common::Registerer<Extractor>::NewInstance(
      "Extractor" + config_["lidar"].as<std::string>()));
  if (!extractor_->Init()) {  // 初始化提取器失败
    AERROR << "Failed to initialize extractor"; // 初始化里程计失败
    return false;
  }
  // 创建里程计实例
  odometer_.reset(new Odometer);
  if (!odometer_->Init()) {
    AERROR << "Failed to initialize odometer";
    return false;
  }
  // 创建地图生成器实例
  mapper_.reset(new Mapper);
  if (!mapper_->Init()) { // 初始化地图生成器失败
    AERROR << "Failed to initialize mapper";
    return false;
  }
  // 如果启用了可视化，初始化可视化器
  if (is_vis_) {
    visualizer_.reset(
        new OhmyloamVisualizer(config_["save_map_path"].as<std::string>()));
  }
  return true;
}

// 如果启用了可视化，初始化可视化器
void OhMyLoam::Reset() {
  AWARN << "OhMySlam RESET"; // 输出重置警告信息
  extractor_->Reset(); // 重置提取器
  odometer_->Reset();  // 重置里程计
  mapper_->Reset();    // 重置地图生成器
}

// 处理每一帧的点云数据
void OhMyLoam::Run(double timestamp,
                   const common::PointCloudConstPtr &cloud_in) {
  common::PointCloudPtr cloud(new common::PointCloud);
  RemoveOutliers(*cloud_in, cloud.get()); // 去除离群点
  std::vector<Feature> features;
  extractor_->Process(timestamp, cloud, &features); // 提取特征
  common::Pose3d pose_curr2odom;
  odometer_->Process(timestamp, features, &pose_curr2odom);  // 处理里程计数据
  common::Pose3d pose_curr2map;
  const auto &cloud_corn = odometer_->GetCloudCorn()->makeShared(); // 获取稀疏点云
  const auto &cloud_surf = odometer_->GetCloudSurf()->makeShared(); // 获取平面点云
  mapper_->Process(timestamp, cloud_corn, cloud_surf, pose_curr2odom,
                   &pose_curr2map); // 处理地图生成
  if (is_vis_) {
    Visualize(pose_curr2map, cloud_corn, cloud_surf, timestamp); // 可视化
  }
}

// 可视化函数
void OhMyLoam::Visualize(const common::Pose3d &pose_curr2map,
                         const TPointCloudConstPtr &cloud_corn,
                         const TPointCloudConstPtr &cloud_surf,
                         double timestamp) {
  std::shared_ptr<OhmyloamVisFrame> frame(new OhmyloamVisFrame); // 创建可视化帧
  frame->timestamp = timestamp;
  frame->cloud_map_corn = mapper_->GetMapCloudCorn();  // 获取当前地图的角点点云
  frame->cloud_map_surf = mapper_->GetMapCloudSurf();  // 获取当前地图的平面点云
  frame->cloud_corn = cloud_corn;   // 当前帧的稀疏点云
  frame->cloud_surf = cloud_surf;   // 当前帧的平面点云
  frame->pose_map = pose_curr2map;  // 当前帧的地图位姿
  visualizer_->Render(frame);       // 渲染可视化帧
}

// 去除点云中的离群点
void OhMyLoam::RemoveOutliers(const common::PointCloud &cloud_in,
                              common::PointCloud *const cloud_out) const {
  common::RemovePoints<common::Point>(
      cloud_in, cloud_out, [&](const common::Point &pt) {
        // 去除不合理的点（非有限值或距离过小的点）
        return !common::IsFinite(pt) ||
               common::DistanceSquare(pt) < kPointMinDist * kPointMinDist;
      });
}

}  // namespace oh_my_loam